/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.1.1)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2006 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* CVS $Id: l_real.hpp,v 1.11 2006/03/08 14:16:27 cxsc Exp $ */

#ifndef L_REAL_INCLUDED
#define L_REAL_INCLUDED

#include <iostream>
#include <string>

#include "dot.hpp"
#include "real.hpp"
#include "interval.hpp" // Blomquist 03.10.02; for _interval(t) in times2pown
#include "except.hpp"

namespace cxsc {

class l_interval;
class interval;

extern int stagprec;

class l_real
{
   private:
      // ---- Data elements ----
      int   prec;
      real  *data;

   public:
      // ---- Constructors ----
      l_real(void) throw();
      l_real(const l_real &) throw();
      

      l_real & operator =(const l_real &) throw();
      l_real & operator =(const real &) throw();
      l_real & operator =(const dotprecision &) throw();

      // ---- Destructors    ----
      ~l_real(void) throw();

      // ---- Type casting ----
      explicit l_real(int)          throw();
      explicit l_real(long)          throw();
      explicit l_real(const real &) throw();
      explicit l_real(const dotprecision &) throw();
 explicit l_real(const double &) throw(); // Blomquist 10.09.02. {l_real.cpp}
      
      friend real::real(const l_real &) throw();
      friend dotprecision::dotprecision(const l_real &) throw();
      friend dotprecision & dotprecision::operator =(const l_real &) throw();

      friend real _real(const l_real & a) throw() { return real(a); }
      friend l_real _l_real(const real & a) throw()  { return l_real(a); }
      friend dotprecision _dotprecision(const l_real & a) throw()
                                               { return dotprecision(a); }
      friend l_real _l_real(const dotprecision & a) throw() 
                                               { return l_real(a); }
      friend l_interval _unchecked_l_interval(const l_real &, 
                                              const l_real &) throw();

      // The following are defined in the specific vector, matrix-files
#if(IndCheck) 
		explicit INLINE l_real(const l_rvector &)       throw (ERROR_LRVECTOR_TYPE_CAST_OF_THICK_OBJ,ERROR_LRVECTOR_USE_OF_UNINITIALIZED_OBJ);
      explicit INLINE l_real(const l_rvector_slice &) throw (ERROR_LRVECTOR_TYPE_CAST_OF_THICK_OBJ,ERROR_LRVECTOR_USE_OF_UNINITIALIZED_OBJ);
      explicit INLINE l_real(const l_rmatrix &)       throw (ERROR_LRMATRIX_TYPE_CAST_OF_THICK_OBJ,ERROR_LRMATRIX_USE_OF_UNINITIALIZED_OBJ);
      explicit INLINE l_real(const l_rmatrix_slice &) throw (ERROR_LRMATRIX_TYPE_CAST_OF_THICK_OBJ,ERROR_LRMATRIX_USE_OF_UNINITIALIZED_OBJ);
      friend INLINE real _l_real(const l_rvector &)       throw (ERROR_LRVECTOR_TYPE_CAST_OF_THICK_OBJ,ERROR_LRVECTOR_USE_OF_UNINITIALIZED_OBJ);
      friend INLINE real _l_real(const l_rvector_slice &) throw (ERROR_LRVECTOR_TYPE_CAST_OF_THICK_OBJ,ERROR_LRVECTOR_USE_OF_UNINITIALIZED_OBJ);
      friend INLINE real _l_real(const l_rmatrix &)       throw (ERROR_LRMATRIX_TYPE_CAST_OF_THICK_OBJ,ERROR_LRMATRIX_USE_OF_UNINITIALIZED_OBJ);
      friend INLINE real _l_real(const l_rmatrix_slice &) throw (ERROR_LRMATRIX_TYPE_CAST_OF_THICK_OBJ,ERROR_LRMATRIX_USE_OF_UNINITIALIZED_OBJ);
#else
		explicit INLINE l_real(const l_rvector &)       throw ();
      explicit INLINE l_real(const l_rvector_slice &) throw ();
      explicit INLINE l_real(const l_rmatrix &)       throw ();
      explicit INLINE l_real(const l_rmatrix_slice &) throw ();
      friend INLINE real _l_real(const l_rvector &)       throw ();
      friend INLINE real _l_real(const l_rvector_slice &) throw ();
      friend INLINE real _l_real(const l_rmatrix &)       throw ();
      friend INLINE real _l_real(const l_rmatrix_slice &) throw ();
#endif


      // ---- Output functions ----
      friend std::ostream & operator <<(std::ostream &,const l_real &) throw();
      friend std::istream & operator >>(std::istream &,l_real &)       throw();
      friend std::string & operator <<(std::string &,const l_real &)   throw();
      friend std::string & operator >>(std::string &,l_real &)         throw();
      friend void          operator >>(const std::string &,l_real &)   throw();
      friend void          operator >>(const char *,l_real &)          throw();

      // ---- Standard functions ---- (arithmetic operators)
      real&             operator[](int) const throw();

      friend     l_real operator -(const l_real& lr1) throw();
      friend     l_real operator +(const l_real& lr1) throw();

      friend     l_real operator +(const l_real &,const l_real &) throw();
      friend     l_real operator -(const l_real &,const l_real &) throw();
      friend     l_real operator *(const l_real &,const l_real &) throw();
      friend     l_real operator /(const l_real &,const l_real &) throw(DIV_BY_ZERO);
      friend inline l_interval operator |(const l_real &,const l_real &) throw();

      friend     l_real operator +(const l_real &,const real &) throw();
      friend     l_real operator +(const real &,const l_real &) throw();
      friend     l_real operator -(const l_real &,const real &) throw();
      friend     l_real operator -(const real &,const l_real &) throw();
      friend     l_real operator *(const l_real &,const real &) throw();
      friend     l_real operator *(const real &,const l_real &) throw();
      friend     l_real operator /(const l_real &,const real &) throw();
      friend     l_real operator /(const real &,const l_real &) throw();
      friend inline l_interval operator |(const real &,const l_real &) throw();
      friend inline l_interval operator |(const l_real &,const real &) throw();

      friend  dotprecision operator +(const l_real &,const dotprecision &) throw();
      friend  dotprecision operator +(const dotprecision &,const l_real &) throw();
      friend  dotprecision operator -(const l_real &,const dotprecision &) throw();
      friend  dotprecision operator -(const dotprecision &,const l_real &) throw();
      friend inline idotprecision operator |(const dotprecision &,const l_real &) throw();
      friend inline idotprecision operator |(const l_real &,const dotprecision &) throw();

      friend     l_real & operator +=(l_real &,const l_real &) throw();
      friend     l_real & operator -=(l_real &,const l_real &) throw();
      friend     l_real & operator *=(l_real &,const l_real &) throw();
      friend     l_real & operator /=(l_real &,const l_real &) throw();

      friend     l_real & operator +=(l_real &,const real &) throw();      
      friend     l_real & operator -=(l_real &,const real &) throw();
      friend     l_real & operator *=(l_real &,const real &) throw();      
      friend     l_real & operator /=(l_real &,const real &) throw();

      friend     real   & operator +=(real &,const l_real &) throw();
      friend     real   & operator -=(real &,const l_real &) throw();
      friend     real   & operator *=(real &,const l_real &) throw();
      friend     real   & operator /=(real &,const l_real &) throw();
      
      friend     inline dotprecision & operator +=(dotprecision &d,const l_real &lr) throw() { lr._akku_add(d); return d; }
      friend     inline dotprecision & operator -=(dotprecision &d,const l_real &lr) throw() { lr._akku_sub(d); return d; }

      // ---- Compare operators ----
      friend bool operator ==(const l_real &,const l_real &) throw();
      friend bool operator !=(const l_real &,const l_real &) throw();
      friend bool operator  <(const l_real &,const l_real &) throw();
      friend bool operator  >(const l_real &,const l_real &) throw();
      friend bool operator <=(const l_real &,const l_real &) throw();
      friend bool operator >=(const l_real &,const l_real &) throw();

      friend bool operator ==(const real &,const l_real &) throw();
      friend bool operator !=(const real &,const l_real &) throw();
      friend bool operator  <(const real &,const l_real &) throw();
      friend bool operator  >(const real &,const l_real &) throw();
      friend bool operator <=(const real &,const l_real &) throw();
      friend bool operator >=(const real &,const l_real &) throw();

      friend bool operator ==(const l_real &,const real &) throw();
      friend bool operator !=(const l_real &,const real &) throw();
      friend bool operator  <(const l_real &,const real &) throw();
      friend bool operator  >(const l_real &,const real &) throw();
      friend bool operator <=(const l_real &,const real &) throw();
      friend bool operator >=(const l_real &,const real &) throw();

      friend bool operator ==(const dotprecision &,const l_real &) throw();
      friend bool operator !=(const dotprecision &,const l_real &) throw();
      friend bool operator  <(const dotprecision &,const l_real &) throw();
      friend bool operator  >(const dotprecision &,const l_real &) throw();
      friend bool operator <=(const dotprecision &,const l_real &) throw();
      friend bool operator >=(const dotprecision &,const l_real &) throw();

      friend bool operator ==(const l_real &,const dotprecision &) throw();
      friend bool operator !=(const l_real &,const dotprecision &) throw();
      friend bool operator  <(const l_real &,const dotprecision &) throw();
      friend bool operator  >(const l_real &,const dotprecision &) throw();
      friend bool operator <=(const l_real &,const dotprecision &) throw();
      friend bool operator >=(const l_real &,const dotprecision &) throw();

      friend bool operator ==(const interval &,const l_real &) throw();
      friend bool operator !=(const interval &,const l_real &) throw();
      friend bool operator  <(const interval &,const l_real &) throw();
      friend bool operator  >(const interval &,const l_real &) throw();
      friend bool operator <=(const interval &,const l_real &) throw();
      friend bool operator >=(const interval &,const l_real &) throw();

      friend bool operator ==(const l_real &,const interval &) throw();
      friend bool operator !=(const l_real &,const interval &) throw();
      friend bool operator  <(const l_real &,const interval &) throw();
      friend bool operator  >(const l_real &,const interval &) throw();
      friend bool operator <=(const l_real &,const interval &) throw();
      friend bool operator >=(const l_real &,const interval &) throw();

      friend bool operator ==(const idotprecision &,const l_real &) throw();
      friend bool operator !=(const idotprecision &,const l_real &) throw();
      friend bool operator  <(const idotprecision &,const l_real &) throw();
      friend bool operator  >(const idotprecision &,const l_real &) throw();
      friend bool operator <=(const idotprecision &,const l_real &) throw();
      friend bool operator >=(const idotprecision &,const l_real &) throw();

      friend bool operator ==(const l_real &,const idotprecision &) throw();
      friend bool operator !=(const l_real &,const idotprecision &) throw();
      friend bool operator  <(const l_real &,const idotprecision &) throw();
      friend bool operator  >(const l_real &,const idotprecision &) throw();
      friend bool operator <=(const l_real &,const idotprecision &) throw();
      friend bool operator >=(const l_real &,const idotprecision &) throw();


      friend bool operator!(const l_real& lr) throw();
      
      // ---- functions ----
      friend void accumulate(dotprecision&, const real&, const l_real&) throw();
      friend void accumulate(dotprecision&, const l_real&, const real&) throw();
      friend void accumulate(dotprecision&, const l_real&, const l_real&) throw();
      
      friend void accumulate(idotprecision&,const real&, const l_real&) throw();
      friend void accumulate(idotprecision&,const l_real&,const real&) throw();
      friend void accumulate(idotprecision&,const l_real&, const l_real&) throw();

      friend l_real   abs  (const l_real&) throw();
      friend int      sign (const l_real&) throw();
      friend int      StagPrec(const l_real&) throw();
      friend l_real   adjust(const l_real&) throw();

      // ---- Friends      -----
      
      friend inline l_real Inf(const l_interval &) throw();
      friend inline l_real Sup(const l_interval &) throw();
      friend        l_real mid(const l_interval &) throw();
    friend inline bool zero_(const l_real &) throw(); // Blomquist,27.11.02
   private:
      void _clear(int) throw(); // filling a l_real number from element int p
                                //  up to the end with zero.
      void _akku_out() throw(); // The dotakku[0] value is roundet to the
                        // activated l_real number in its own precision.
      void _akku_add(dotprecision&) const throw(); // adding the activated
                                     // l_real number to the accumulator d.
      void _akku_sub(dotprecision&) const throw(); // subtracting the 
      // activated l_real number to the accumulator d of type dotprecision.
      inline real& elem(int i) const throw() {  return data[i-1];  }
};

    inline l_real l_pow2n(const int n) throw()
{   // Fast and exact calculation of 2^n; -1074 <= n <= 1023;
    // Blomquist 01.10.02.
    return l_real( comp(0.5,n+1) );
}

inline void times2pown(l_real& lr, const int n) throw() // Blomquist 03.10.02
{ // lr is multiplied with 2^n; if lr[i]*2^n are all normalized, the result
  // is exact. if one of the lr[i]*2^n are denormalized, the result is not
  // exact in general.
    int k = StagPrec(lr);
    for (int i=1; i<=k; i++)
    {
	times2pown(lr[i],n);
    }
}

inline void times2pown(l_real& lr, interval& z, const int n) throw() 
{  // Blomquist 03.10.02;
// If we denote the old value of lr with y, then times2pown(lr,z,n)
// delivers with lr+z an inclusion of the exact value y*2^n;
// z=0 signals that  lr = y *2^n  was exactly calculated.
    if ( n<-1074 || n>1023 ) 
    { std::cerr << "Error in:  " 
           << "times2pown(l_real& lr, interval& z, const int n): " << std::endl
           << " -1074 <= n <= +1023 not fulfilled" << std::endl; exit(0); } 
    int k = StagPrec(lr);
    z = 0;
    real mt,t;
    real F = comp(0.5,n+1);
    for (int i=1; i<=k; i++)
    {
	mt = mant(lr[i]);
	t = lr[i];
	times2pown(lr[i],n);
	if ( mt != mant(lr[i]) ) 
	{
	    lr[i] = 0;
	    z += _interval(t) * F;
        }
    }

}

inline bool zero_(const l_real& lr) throw()
{  // returns only true if all lr.elem(i) == 0; Blomquist, 27.11.02; 
    int i=1, p=StagPrec(lr);
    bool tmp = true;
    do
    {
	if (sign(lr.elem(i))!=0) tmp = false;
	i++;
    }  while(tmp && i <= p );
    return tmp;
}

} // namespace cxsc 

#endif
